#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -*- Python -*-
import json
import logging
import paho.mqtt.client as mqtt
import threading
import time
import ssl
import datetime


logger = logging.getLogger("VIEWER").getChild(__name__)


class MqttInterface(mqtt.Client):

    def __init__(self, set_host, set_port, sub_topic="", mqtt_queue=None, location=""):
        super(MqttInterface, self).__init__()
        self._sub_topic = sub_topic
        self._host = set_host
        self._port = set_port
        self._sub_connect_flag = False
        self._sub_receive_flag = False
        self._sub_receive_message = None
        self._lock = threading.Lock()
        self._RETRY_CONNECT_MAX = 3
        self._CONNECT_WAIT_MAX = 3
        self._CONNECT_WAIT = 1
        self.message_queue = mqtt_queue
        self.location = location

    def connection(self, ca_path="", client_path="", client_key_path=""):
        """
        Connect to MQTT Server

        Returns
        ----------
        result: 接続結果
        """
        # Connection initial setting
        result = False
        self._sub_connect_flag = False
        retry_count = 0
        try:
            # Retry processing
            while not self._sub_connect_flag and retry_count < self._RETRY_CONNECT_MAX:
                self.tls_set(ca_path, certfile=client_path, keyfile=client_key_path, cert_reqs=mqtt.ssl.CERT_REQUIRED, tls_version=ssl.PROTOCOL_TLSv1_2)
                # Connection
                self.connect(self._host, self._port, 60)
                # Perform subscription processing when there is a topic
                if self._sub_topic:
                    self.subscribe(self._sub_topic, qos=0)
                # loop start processing
                self.loop_start()
                # Connection result confirmation process
                for wait_count in range(self._CONNECT_WAIT_MAX):
                    time.sleep(self._CONNECT_WAIT)
                    if self._sub_connect_flag:
                        result = True
                        break
                else:
                    # Retry count addition
                    retry_count += 1
        except Exception as e:
            logger.error(e)
        if not result:
            logger.error("Unable to connection for MQTT server")
        return result

    def on_connect(self, client, userdata, flag, rc):
        """
        MQTT サーバーとの接続が確立されたときに呼び出されるコールバック

        Parameters
        ----------
        client: このコールバックのクライアント インスタンス
        userdata: プライベート ユーザー データ
        flag: ブローカーから送信された応答フラグ
        rc: 接続結果
        """
        logger.info("Connected with result code {}".format(rc))
        if rc == mqtt.MQTT_ERR_SUCCESS:
            self._sub_connect_flag = True
        else:
            self._sub_connect_flag = False

    def on_message(self, client, userdata, msg):
        """
        対象Topicのメッセージを受信した際に呼び出されるコールバック

        Parameters
        ----------
        client: クライアントインスタンス
        userdata: ユーザデータ
        msg: MQTTMessage インスタンス
        """
        # Convert message from Json to dictionary
        logger.info("[{}] call message".format(datetime.datetime.now()))
        dict_message = json.loads(msg.payload)
        logger.debug("message: {}".format(dict_message))
        self._lock.acquire()
        self._sub_receive_flag = True
        self._sub_receive_message = dict_message
        if self.message_queue is None:
            self._lock.release()
            return
        logger.debug(dict_message)
        location = dict_message.get("location")
        location_name = location.get("name")
        if location_name.lower() == self.location.lower():
            print("\n[DEBUG]on_message:[{}]\n".format(dict_message))
            self.message_queue.put(dict_message, block=True, timeout=None)
        self._lock.release()

