import multiprocessing
import os
import datetime
import logging
import json
import queue
import time
import requests

from mqtt_interface import MqttInterface
from create_yaml import CreateYaml
from rearrangement_obstacles import *
from socket_server import start_server


FORMAT = "%(asctime)s-%(levelname)s: %(message)s :fn[%(module)s:%(funcName)s]"
LOG_TIME = datetime.datetime.now(datetime.timezone(datetime.timedelta(hours=9), "JST"))
LOG_DIR_PATH = os.path.join(os.path.dirname(__file__), "./log")
LOG_FILE = LOG_DIR_PATH + "/log_" + LOG_TIME.strftime("%Y%m%d%H%M%S") + ".log"

if not os.path.exists(LOG_DIR_PATH):
    os.mkdir(LOG_DIR_PATH)

logger = logging.getLogger("VIEWER")
logger.setLevel(logging.DEBUG)
handler = logging.FileHandler(LOG_FILE)
handler.setLevel(logging.DEBUG)
formatter = logging.Formatter(FORMAT)
handler.setFormatter(formatter)
logger.addHandler(handler)

MQTT_TOPIC = "obj_recognition_01/estimated_position"
# MQTT_CONF_PATH = os.path.join(os.path.dirname(__file__), "./mqtt_conf.json")
MQTT_CONF_PATH = os.path.join(os.path.dirname(__file__), "./mqtt_conf_local.json")
LOCATION_CONF_PATH = os.path.join(os.path.dirname(__file__), "./location_conf.json")

# GET_LAYOUT_API_URL = "https://uoa-rdr.jp/RDBS/data/rdr/get/LayoutInfo"
GET_LAYOUT_API_URL = "http://localhost:8080/RDBS/data/rdr/get/Layout"


class ObjectViewer:

    mqtt_ins = None
    mqtt_queue = queue.Queue()
    base_yaml = None
    location_name = ""
    semi_static_body = []
    obstacles_body = []

    @staticmethod
    def json_read(path):
        """
        Jsonファイルを読み込む
        :param path:ファイルパス
        :return 1: 読み込み結果
        :return 2: 読み込んだデータ
        """
        load_data = None
        result = False
        try:
            if not os.path.exists(path):
                return result, load_data
            with open(path, "r") as conf_file:
                load_data = json.load(conf_file)
                result = True
        except json.JSONDecodeError as err:
            logger.error("Unable to decode json file")
        except Exception as err:
            logger.error("Unable to read setting file")
        return result, load_data

    @classmethod
    def location_setting(cls):
        """
        ロケーション設定処理

        :return: 設定結果
        """
        result = False

        location_res, location_config = cls.json_read(LOCATION_CONF_PATH)
        if not location_res:
            logger.error("Unable to read location config file")
            return result

        base_yaml = CreateYaml.cnv_config_2_yaml(location_config)
        if base_yaml is None:
            return result
        cls.location_name = location_config.get("location")
        cls.base_yaml = base_yaml
        cls.semi_static_body = location_config.get("semi_static_bodys")
        cls.obstacles_body = location_config.get("obstacle_bodys")
        result = True
        return result

    @classmethod
    def mqtt_setting(cls):
        """
        MQTT接続設定処理

        :return: 設定結果
        """
        result = False

        mqtt_res, mqtt_config = cls.json_read(MQTT_CONF_PATH)
        if not mqtt_res:
            logger.error("Unable to read mqtt config file")
            return result

        cls.mqtt_ins = MqttInterface(mqtt_config["host"], mqtt_config["port"], sub_topic=MQTT_TOPIC, mqtt_queue=cls.mqtt_queue, location=cls.location_name)
        if cls.mqtt_ins.connection(mqtt_config["ca_certs"], mqtt_config["cert_file"], mqtt_config["key_file"]):
            result = True

        return result

    @staticmethod
    def request_get_layout(location):
        """
        レイアウトデータをサーバーから取得する

        Returns
        ----------
        result: 送信結果
        result: レイアウトList
        """
        logger.debug("call layout get req")
        result = False
        layout_data = []

        try:
            post_res = requests.post(GET_LAYOUT_API_URL, json={"location": location}, timeout=3.0)
            if post_res.status_code == requests.status_codes.codes.ok or \
                    post_res.status_code == requests.status_codes.codes.NO_CONTENT:
                if post_res.status_code == requests.status_codes.codes.ok:
                    layout_dict = json.loads(post_res.text)
                    layout_data = layout_dict.get("layout_list")
                result = True
            else:
                pass
        except json.JSONDecodeError as e:
            logger.error(e)
        except Exception as e:
            logger.error(e)
        return result, layout_data

    @classmethod
    def initial_placement(cls):
        """
        Choreonoidの初期配置を行う
        :return:
        """
        cnoid_ins = RearrangementObstacles(cls.base_yaml)

        if cnoid_ins.get_world_item() is not None:
            return

        cnoid_ins.scene_view_properties()
        # Generate world item.
        cnoid_ins.world_item_generator()

        # Read environment models.
        cnoid_ins.read_environment_model()


    @classmethod
    def run(cls, active_time_memory: multiprocessing.Value):
        """
        物体配置処理
        :param active_time_memory: 処理実行判断共有メモリ
        :return:
        """
        try:
            cls.mqtt_queue_clear()
            cnoid_ins = RearrangementObstacles(cls.base_yaml)
            semi_static_object_removal_time = None
            while active_time_memory.value:
                if not cls.mqtt_queue.empty():
                    current_time = datetime.datetime.now()
                    view_object_a = [body.get("name") for body in cls.obstacles_body]
                    object_data = cls.mqtt_queue.get(block=True, timeout=None)
                    object_a = object_data.get("object_list")
                    reference_body_a = copy.deepcopy(cls.obstacles_body)
                    if object_a is None:
                        cls.mqtt_queue.task_done()
                        continue

                    if semi_static_object_removal_time is None or semi_static_object_removal_time < current_time:
                        # レイアウト情報取得
                        get_res, location_data = cls.request_get_layout(cls.location_name)
                        logger.info("call layout : {}".format(semi_static_object_removal_time))
                        semi_static_object_removal_time = current_time + datetime.timedelta(seconds=20)

                        if get_res:

                            view_object_a += [body.get("name") for body in cls.semi_static_body]
                            reference_body_a += cls.semi_static_body
                            object_a += location_data
                    # 表示データ更新
                    update_res, update_yaml = CreateYaml.update(object_a, cls.base_yaml, reference_body_a)

                    if not update_res:
                        logger.error("Unable to update")
                        cls.mqtt_queue.task_done()
                        continue
                    cnoid_ins.reallocate_free_space(update_yaml, view_object_a)

                    cls.mqtt_queue.task_done()
                time.sleep(0.1)
        except KeyboardInterrupt as err:
            logger.info(err)
        except Exception as err:
            logger.error(err)



    @classmethod
    def mqtt_queue_clear(cls):
        while not cls.mqtt_queue.empty():
            cls.mqtt_queue.get()


def main():
    object_viewer_ins = ObjectViewer()

    if not object_viewer_ins.location_setting():
        return
    if not object_viewer_ins.mqtt_setting():
        return
    # 共有メモリ宣言
    shared_memory = multiprocessing.Value('b', True)
    active_time_memory = multiprocessing.Value('b', True)

    # サーバーを開始
    start_server(shared_memory, active_time_memory)

    try:
        while True:
            if active_time_memory.value:
                object_viewer_ins.run(active_time_memory)
            if not shared_memory.value:
                object_viewer_ins.mqtt_ins.disconnect()
                break
            time.sleep(1)
    except Exception as e:
        logger.error(e)



if __name__ == "__main__":
    main()
